#include <SPI.h>
#include <avr/pgmspace.h>
#include <Mouse.h>
#include <Keyboard.h>
#include "Rotary.h"

// Registers
#define Product_ID  0x00
#define Revision_ID 0x01
#define Motion  0x02
#define Delta_X_L 0x03
#define Delta_X_H 0x04
#define Delta_Y_L 0x05
#define Delta_Y_H 0x06
#define SQUAL 0x07
#define Raw_Data_Sum  0x08
#define Maximum_Raw_data  0x09
#define Minimum_Raw_data  0x0A
#define Shutter_Lower 0x0B
#define Shutter_Upper 0x0C
#define Control 0x0D
#define Config1 0x0F
#define Config2 0x10
#define Angle_Tune  0x11
#define Frame_Capture 0x12
#define SROM_Enable 0x13
#define Run_Downshift 0x14
#define Rest1_Rate_Lower  0x15
#define Rest1_Rate_Upper  0x16
#define Rest1_Downshift 0x17
#define Rest2_Rate_Lower  0x18
#define Rest2_Rate_Upper  0x19
#define Rest2_Downshift 0x1A
#define Rest3_Rate_Lower  0x1B
#define Rest3_Rate_Upper  0x1C
#define Observation 0x24
#define Data_Out_Lower  0x25
#define Data_Out_Upper  0x26
#define Raw_Data_Dump 0x29
#define SROM_ID 0x2A
#define Min_SQ_Run  0x2B
#define Raw_Data_Threshold  0x2C
#define Config5 0x2F
#define Power_Up_Reset  0x3A
#define Shutdown  0x3B
#define Inverse_Product_ID  0x3F
#define LiftCutoff_Tune3  0x41
#define Angle_Snap  0x42
#define LiftCutoff_Tune1  0x4A
#define Motion_Burst  0x50
#define LiftCutoff_Tune_Timeout 0x58
#define LiftCutoff_Tune_Min_Length  0x5A
#define SROM_Load_Burst 0x62
#define Lift_Config 0x63
#define Raw_Data_Burst  0x64
#define LiftCutoff_Tune2  0x65

//Set this to what pin your "INT0" hardware interrupt feature is on
#define Motion_Interrupt_Pin 5

const int ncs = 0;  //This is the SPI "slave select" pin that the sensor is hooked up to

byte initComplete = 0;
volatile int xydat[2];
volatile byte movementflag = 0;


//Be sure to add the SROM file into this sketch via "Sketch->Add File"
extern const unsigned short firmware_length;
extern const unsigned char firmware_data[];

byte x = 0;
byte y = 0;

Rotary rotary = Rotary(5, 2);

const int buttonPin2 = 4;     // the number of the pushbutton pin
const int buttonPin3 = 16;     // the number of the pushbutton pin   left

const int buttonPin4 = 6;     // the number of the pushbutton pin
const int buttonPin5 = 15;     // the number of the pushbutton pin   right

const int buttonPin6 = 7;     // the number of the pushbutton pin
const int buttonPin7 = 14;     // the number of the pushbutton pin    middle



const int ledPin1 =  22;      // the number of the LED pin
const int ledPin2 =  23;      // the number of the LED pin
const int ledPin3 =  13;      // the number of the LED pin


// variables will change:
int buttonState2 = 0;         // variable for reading the pushbutton status
int buttonState3 = 0;         // variable for reading the pushbutton status

int buttonState4 = 0;         // variable for reading the pushbutton status
int buttonState5 = 0;         // variable for reading the pushbutton status

int buttonState6 = 0;         // variable for reading the pushbutton status
int buttonState7 = 0;         // variable for reading the pushbutton status


void setup() {
  Serial.begin(9600);

  pinMode (ncs, OUTPUT);

  pinMode(Motion_Interrupt_Pin, INPUT);
  digitalWrite(Motion_Interrupt_Pin, HIGH);
  attachInterrupt(0, UpdatePointer, FALLING);

  SPI.begin();
  SPI.setDataMode(SPI_MODE3);
  SPI.setBitOrder(MSBFIRST);
  //SPI.setClockDivider(4);


  performStartup();

  delay(5000);

  dispRegisters();
  initComplete = 9;

  Mouse.begin(); //start mouse emulation
  pinMode(buttonPin2, INPUT_PULLUP); // +click
  pinMode(buttonPin3, INPUT_PULLUP); // -click

  pinMode(buttonPin4, INPUT_PULLUP); // +click
  pinMode(buttonPin5, INPUT_PULLUP); // -click

  pinMode(buttonPin6, INPUT_PULLUP); // +click
  pinMode(buttonPin7, INPUT_PULLUP); // -click

}

void adns_com_begin() {
  digitalWrite(ncs, LOW);
}

void adns_com_end() {
  digitalWrite(ncs, HIGH);
}

byte adns_read_reg(byte reg_addr) {
  adns_com_begin();

  // send adress of the register, with MSBit = 0 to indicate it's a read
  SPI.transfer(reg_addr & 0x7f );
  delayMicroseconds(100); // tSRAD
  // read data
  byte data = SPI.transfer(0);

  delayMicroseconds(1); // tSCLK-NCS for read operation is 120ns
  adns_com_end();
  delayMicroseconds(19); //  tSRW/tSRR (=20us) minus tSCLK-NCS

  return data;
}

void adns_write_reg(byte reg_addr, byte data) {
  adns_com_begin();

  //send adress of the register, with MSBit = 1 to indicate it's a write
  SPI.transfer(reg_addr | 0x80 );
  //sent data
  SPI.transfer(data);

  delayMicroseconds(20); // tSCLK-NCS for write operation
  adns_com_end();
  delayMicroseconds(100); // tSWW/tSWR (=120us) minus tSCLK-NCS. Could be shortened, but is looks like a safe lower bound
}

void adns_upload_firmware() {
  // send the firmware to the chip, cf p.18 of the datasheet
  Serial.println("Uploading firmware...");

  //Write 0 to Rest_En bit of Config2 register to disable Rest mode.
  adns_write_reg(Config2, 0x20);

  // write 0x1d in SROM_enable reg for initializing
  adns_write_reg(SROM_Enable, 0x1d);

  // wait for more than one frame period
  delay(10); // assume that the frame rate is as low as 100fps... even if it should never be that low

  // write 0x18 to SROM_enable to start SROM download
  adns_write_reg(SROM_Enable, 0x18);

  // write the SROM file (=firmware data)
  adns_com_begin();
  SPI.transfer(SROM_Load_Burst | 0x80); // write burst destination adress
  delayMicroseconds(15);

  // send all bytes of the firmware
  unsigned char c;
  for (int i = 0; i < firmware_length; i++) {
    c = (unsigned char)pgm_read_byte(firmware_data + i);
    SPI.transfer(c);
    delayMicroseconds(15);
  }

  //Read the SROM_ID register to verify the ID before any other register reads or writes.
  adns_read_reg(SROM_ID);

  //Write 0x00 to Config2 register for wired mouse or 0x20 for wireless mouse design.
  adns_write_reg(Config2, 0x00);

  // set initial CPI resolution
  adns_write_reg(Config1, 0x15);

  adns_com_end();
}


void performStartup(void) {
  adns_com_end(); // ensure that the serial port is reset
  adns_com_begin(); // ensure that the serial port is reset
  adns_com_end(); // ensure that the serial port is reset
  adns_write_reg(Power_Up_Reset, 0x5a); // force reset
  delay(50); // wait for it to reboot
  // read registers 0x02 to 0x06 (and discard the data)
  adns_read_reg(Motion);
  adns_read_reg(Delta_X_L);
  adns_read_reg(Delta_X_H);
  adns_read_reg(Delta_Y_L);
  adns_read_reg(Delta_Y_H);
  // upload the firmware
  adns_upload_firmware();
  delay(10);
  Serial.println("Optical Chip Initialized");
}

void UpdatePointer(void) {
  if (initComplete == 9) {

    //write 0x01 to Motion register and read from it to freeze the motion values and make them available
    adns_write_reg(Motion, 0x01);
    adns_read_reg(Motion);

    xydat[0] = (int)adns_read_reg(Delta_X_L);
    xydat[1] = (int)adns_read_reg(Delta_Y_L);

    movementflag = 1;
  }
}

void dispRegisters(void) {
  int oreg[7] = {
    0x00, 0x3F, 0x2A, 0x02
  };
  char* oregname[] = {
    "Product_ID", "Inverse_Product_ID", "SROM_Version", "Motion"
  };
  byte regres;

  digitalWrite(ncs, LOW);

  int rctr = 0;
  for (rctr = 0; rctr < 4; rctr++) {
    SPI.transfer(oreg[rctr]);
    delay(1);
    Serial.println("---");
    Serial.println(oregname[rctr]);
    Serial.println(oreg[rctr], HEX);
    regres = SPI.transfer(0);
    Serial.println(regres, BIN);
    Serial.println(regres, HEX);
    delay(1);
  }
  digitalWrite(ncs, HIGH);
}


int convTwosComp(int b) {
  //Convert from 2's complement
  if (b & 0x80) {
    b = -1 * ((b ^ 0xff) + 1);
  }
  return b;
}

//////////////////////////////////////////

int getX() { //returns the X acceleration value
  byte x = 0;
  x = xydat[0];
  return (convTwosComp(x));
}

int getY() { //returns the Y acceleration value
  byte y = 0;
  y = xydat[1];
  return (convTwosComp(y));
}

/////////////////////////////////////////

void loop() {

  if (movementflag) {

    //uncomment below if using Teensy from PJRC to move mouse on screen
    Mouse.move(convTwosComp(xydat[0]),convTwosComp(xydat[1]));

    Serial.print("x = ");
    Serial.print( convTwosComp(xydat[0]) );
    Serial.print(" | ");
    Serial.print("y = ");
    Serial.println( convTwosComp(xydat[1]) );

    movementflag = 0;
  }

  x = 0 - getX();
  y = getY();
  Mouse.move(x, y, 0); //Move the cursor

  unsigned char result = rotary.process();
  if (result == DIR_CW) {
    x = 0 - getX();
    y = getY();
    Mouse.move(x, y, -1);
  } else if (result == DIR_CCW) {
    x = 0 - getX();
    y = getY();
    Mouse.move(x, y, 1);
  }

  //////////////////////////////////////////
  // read the state of the pushbutton value:
  buttonState2 = digitalRead(buttonPin2);
  buttonState3 = digitalRead(buttonPin3);

  // check if the pushbutton is pressed.
  // if it is, the buttonState is HIGH:

  if (buttonState2 == LOW) {
    // turn LED on:
    //digitalWrite(ledPin, LOW);
    Mouse.press(MOUSE_LEFT);
  }

  if (buttonState3 == LOW) {
    // turn LED off:
    //digitalWrite(ledPin, HIGH);
    Mouse.release(MOUSE_LEFT);
  }
  ///////////////////////////////////////////////
  // read the state of the pushbutton value:
  buttonState4 = digitalRead(buttonPin4);
  buttonState5 = digitalRead(buttonPin5);

  // check if the pushbutton is pressed.
  // if it is, the buttonState is HIGH:

  if (buttonState4 == LOW) {
    // turn LED on:
    //digitalWrite(ledPin, LOW);
    Mouse.press(MOUSE_RIGHT);
  }

  if (buttonState5 == LOW) {
    // turn LED off:
    //digitalWrite(ledPin, HIGH);
    Mouse.release(MOUSE_RIGHT);
  }
  /////////////////////////////////////////////////////////
  // read the state of the pushbutton value:
  buttonState6 = digitalRead(buttonPin6);
  buttonState7 = digitalRead(buttonPin7);

  // check if the pushbutton is pressed.
  // if it is, the buttonState is HIGH:

  if (buttonState6 == LOW) {
    // turn LED on:
    //digitalWrite(ledPin, LOW);
    Mouse.press(MOUSE_MIDDLE);
  }

  if (buttonState7 == LOW) {
    // turn LED off:
    //digitalWrite(ledPin, HIGH);
    Mouse.release(MOUSE_MIDDLE);
  }
  ///////////////////////////////////////////////////
}

